theory IO_Automaton
imports Transition_System Transitions
begin


locale IO_Automaton =
fixes istate :: "'state" (* initial step *)
  and step :: "'state \<Rightarrow> 'act \<Rightarrow> 'out * 'state"  (* transition function *)
begin


section{* Reachability and Invariance *}

(* Reachable states: *)
inductive reach :: "'state \<Rightarrow> bool" where
Istate: "reach istate"
|
Step: "reach s \<Longrightarrow> reach (snd (step s a))"

lemma reach_PairI:
assumes "reach s" and "step s a = (ou, s')"
shows "reach s'"
by (metis Step assms assms snd_eqD)

(* holds at the initial state: *)
definition holdsIstate :: "('state \<Rightarrow> bool) \<Rightarrow> bool" where
"holdsIstate \<phi> \<equiv> \<phi> istate"

(* is invariant: *)
definition invar :: "('state \<Rightarrow> bool) \<Rightarrow> bool" where
"invar \<phi> \<equiv> \<forall> s a. reach s \<and> \<phi> s \<longrightarrow> \<phi> (snd (step s a))"

lemma holdsIstate_invar:
assumes h: "holdsIstate \<phi>" and i: "invar \<phi>" and a: "reach s"
shows "\<phi> s"
using a apply (induct rule: reach.induct)
using h i unfolding holdsIstate_def invar_def by auto


section{* Traces *}

(* The output and effect of an action on a state: *)
definition out :: "'state \<Rightarrow> 'act \<Rightarrow> 'out" where "out s a \<equiv> fst (step s a)"
definition eff :: "'state \<Rightarrow> 'act \<Rightarrow> 'state" where "eff s a \<equiv> snd (step s a)"

fun validTrans :: "('state,'act,'out) trans \<Rightarrow> bool" where
"validTrans (Trans s a ou s') = (step s a = (ou, s'))"

lemma validTrans:
"validTrans trn =
 (step (srcOf trn) (actOf trn) = (outOf trn, tgtOf trn))"
by (cases trn) auto

(* Traces allowed by the system (starting in any given state) *)

(* Two alternative definitions: growing from the left and growing from the right: *)
(* NB: The theories Transition and Transition_System both define (different) type_synonyms "trace".
   The following definitions work because Isabelle seems to use the definition that was loaded
   *last*.

   TODO: Clean up the redundancy between IO automata and transition systems. *)
inductive valid :: "('state,'act,'out) trace \<Rightarrow> bool" where
Singl[simp,intro!]:
"validTrans trn
 \<Longrightarrow>
 valid [trn]"
|
Cons[intro]:
"\<lbrakk>validTrans trn; tgtOf trn = srcOf (hd tr); valid tr\<rbrakk>
 \<Longrightarrow>
 valid (trn # tr)"

inductive_cases valid_SinglE[elim!]: "valid [trn]"
inductive_cases valid_ConsE[elim]: "valid (trn # tr)"


inductive valid2 :: "('state,'act,'out) trace \<Rightarrow> bool" where
Singl[simp,intro!]:
"validTrans trn
 \<Longrightarrow>
 valid2 [trn]"
|
Rcons[intro]:
"\<lbrakk>valid2 tr; tgtOf (last tr) = srcOf trn; validTrans trn\<rbrakk>
 \<Longrightarrow>
 valid2 (tr ## trn)"

inductive_cases valid2_SinglE[elim!]: "valid2 [trn]"
inductive_cases valid2_RconsE[elim]: "valid2 (tr ## trn)"

lemma Nil_not_valid[simp]: "\<not> valid []"
by (metis valid.simps neq_Nil_conv)

lemma Nil_not_valid2[simp]: "\<not> valid2 []"
by (metis valid2.cases append_Nil butlast.simps butlast_snoc not_Cons_self2)

lemma valid_Rcons:
assumes "valid tr" and "tgtOf (last tr) = srcOf trn" and "validTrans trn"
shows "valid (tr ## trn)"
using assms proof(induct arbitrary: trn)
  case (Cons trn tr trna)
  thus ?case by (cases tr) (auto intro: valid.Cons)
qed(auto intro: valid.intros)

lemma valid_hd_Rcons[simp]:
assumes "valid tr"
shows "hd (tr ## tran) = hd tr"
by (metis Nil_not_valid assms hd_append)

lemma valid2_hd_Rcons[simp]:
assumes "valid2 tr"
shows "hd (tr ## tran) = hd tr"
by (metis Nil_not_valid2 assms hd_append)

lemma valid2_last_Cons[simp]:
assumes "valid2 tr"
shows "last (tran # tr) = last tr"
by (metis Nil_not_valid2 assms last.simps)

lemma valid2_Cons:
assumes "valid2 tr" and "tgtOf trn = srcOf (hd tr)" and "validTrans trn"
shows "valid2 (trn # tr)"
using assms proof(induct arbitrary: trn)
  case Singl  show ?case
  unfolding two_singl_Rcons apply(rule valid2.Rcons) using Singl
  by (auto intro: valid2.Singl)
next
  case Rcons show ?case
  unfolding append.append_Cons[symmetric] apply(rule valid2.Rcons) using Rcons by auto
qed

lemma valid_valid2: "valid = valid2"
proof(rule ext, safe)
  fix tr assume "valid tr"  thus "valid2 tr"
  by (induct) (auto intro: valid2.Singl valid2_Cons)
next
  fix tr assume "valid2 tr"  thus "valid tr"
  by (induct) (auto intro: valid.Singl valid_Rcons)
qed

lemma valid_Cons_iff: 
"valid (trn # tr) \<longleftrightarrow> validTrans trn \<and> ((tgtOf trn = srcOf (hd tr) \<and> valid tr) \<or> tr = [])"
unfolding valid.simps[of "trn # tr"] by auto

lemma valid_append: 
"tr \<noteq> [] \<Longrightarrow> tr1 \<noteq> [] \<Longrightarrow> 
 valid (tr @ tr1) \<longleftrightarrow> valid tr \<and> valid tr1 \<and> tgtOf (last tr) = srcOf (hd tr1)"
by (induct tr) (auto simp add: valid_Cons_iff)
 

lemmas valid2_valid = valid_valid2[symmetric]

(* validFrom includes empty traces too (unlike valid): *)
definition validFrom :: "'state \<Rightarrow> ('state,'act,'out) trace \<Rightarrow> bool" where
"validFrom s tr \<equiv> tr = [] \<or> (valid tr \<and> srcOf (hd tr) = s)"

lemma validFrom_Nil[simp,intro!]: "validFrom s []"
unfolding validFrom_def by auto

lemma validFrom_valid[simp,intro]: "valid tr \<and> srcOf (hd tr) = s \<Longrightarrow> validFrom s tr"
unfolding validFrom_def by auto

lemma validFrom_validTrans[intro]:
assumes "validTrans (Trans s a ou s')" and "validFrom s' tr"
shows "validFrom s (Trans s a ou s' # tr)"
using assms unfolding validFrom_def by auto

lemma validFrom_append: 
"validFrom s (tr @ tr1) \<longleftrightarrow> (tr = [] \<and> validFrom s tr1) \<or> (tr \<noteq> [] \<and> validFrom s tr \<and> validFrom (tgtOf (last tr)) tr1)"
unfolding validFrom_def using valid_append 
by (cases "tr = [] \<or> tr1 = []") fastforce+

lemma validFrom_Cons: 
"validFrom s (trn # tr) \<longleftrightarrow> validTrans trn \<and> srcOf trn = s \<and> validFrom (tgtOf trn) tr"
unfolding validFrom_def by auto

(* traces versus reachability: *)
lemma valid_reach_src_tgt:
assumes "valid tr" and "reach (srcOf (hd tr))"
shows "reach (tgtOf (last tr))"
using assms by induct (auto intro: reach_PairI simp: validTrans)

lemma valid_init_reach:
assumes "valid tr" and "srcOf (hd tr) = istate"
shows "reach (tgtOf (last tr))"
using valid_reach_src_tgt assms reach.Istate by metis

(*
lemma validFrom_init_reach:
assumes "validFrom s tr" and "init s"
shows "reach s"
using valid_init_reach assms unfolding validFrom_def by(auto intro: reach.Istate)
*)

lemma Trans_fst_sndI:
"valid [Trans s a (fst (step s a)) (snd (step s a))]"
by (metis valid.Singl surjective_pairing validTrans.simps)

lemma reach_init_valid:
assumes "reach s"
shows
"s = istate
 \<or>
 (\<exists> tr. valid tr \<and> srcOf (hd tr) = istate \<and> tgtOf (last tr) = s)"
using assms proof induction
  case (Step s a)
  thus ?case proof(elim disjE exE conjE)
    assume s: "s = istate"
    let ?ou = "fst (step s a)"  let ?s' = "snd (step s a)"
    show ?thesis
    apply (intro disjI2 exI[of _ "[Trans s a ?ou ?s']"])
    using s by auto
  next
    fix tr assume v: "valid tr" and s: "srcOf (hd tr) = istate" and t: "tgtOf (last tr) = s"
    let ?ou = "fst (step s a)"  let ?s' = "snd (step s a)"
    show ?thesis
    apply (intro disjI2 exI[of _ "tr ## Trans s a ?ou ?s'"])
    using v t s by (auto intro: valid_Rcons)
  qed
qed auto

lemma reach_validFrom:
assumes "reach s'"
shows "\<exists> s tr. s = istate \<and> (s = s' \<or> (validFrom s tr \<and> tgtOf (last tr) = s'))"
using reach_init_valid[OF assms] unfolding validFrom_def by auto

(* The trace given by a state and a ste of actions: *)
fun traceOf :: "'state \<Rightarrow> 'act list \<Rightarrow> ('state,'act,'out)trace" where 
"traceOf s [] = []"
|
"traceOf s (a # al) = 
 (case step s a of (ou,s1) \<Rightarrow> (Trans s a ou s1) # traceOf s1 al)"


(* Multi-step: *)
fun sstep :: "'state \<Rightarrow> 'act list \<Rightarrow> 'out list \<times> 'state" where 
"sstep s [] = ([], s)"
|
"sstep s (a # al) = (case step s a of (ou,s') \<Rightarrow> (case sstep s' al of (oul, s'') \<Rightarrow> (ou # oul, s'')))"

lemma length_traceOf[simp]:
"length (traceOf s al) = length al"
by (induct al arbitrary: s) (auto split: prod.splits)

lemma traceOf_Nil[simp]:
"traceOf s al = [] \<longleftrightarrow> al = []"
by (metis length_traceOf length_0_conv) 

lemma sstep_outOf_traceOf[simp]:
"sstep s al = (ou,s') \<Longrightarrow> map outOf (traceOf s al) = ou"
by (induct al arbitrary: s ou s') (auto split: prod.splits) 

lemma sstep_tgtOf_traceOf[simp]:
"al \<noteq> [] \<Longrightarrow> sstep s al = (ou,s') \<Longrightarrow> tgtOf (last (traceOf s al)) = s'"
by (induct al arbitrary: s ou s') (auto split: prod.splits)

lemma srcOf_traceOf[simp]:
"al \<noteq> [] \<Longrightarrow> srcOf (hd (traceOf s al)) = s"
by (induct al arbitrary: s) (auto split: prod.splits)

lemma actOf_traceOf[simp]:
"map actOf (traceOf s al) = al"
by (induct al arbitrary: s) (auto split: prod.splits)


lemma traceOf_append:
"al \<noteq> [] \<Longrightarrow> s1 = tgtOf (last (traceOf s al)) \<Longrightarrow> 
 traceOf s (al @ al1) = traceOf s al @ traceOf s1 al1"
by (induct al arbitrary: s s1 al1) (auto split: prod.splits)

lemma sstep_append: 
assumes "sstep s al = (oul,s1)" and "sstep s1 al1 = (oul1,s2)"
shows "sstep s (al @ al1) = (oul @ oul1, s2)"
using assms by (induct al arbitrary: oul s s1 oul1 s2) (auto split: prod.splits)

lemma reach_sstep:
assumes "reach s" and "sstep s al = (ou,s1)"
shows "reach s1"
using assms apply(induction al arbitrary: ou s1 s)
by (auto split: prod.splits) (metis reach_PairI)

(* Alternative simp rules, from the left *)
lemma traceOf_consR[simp]: 
assumes "al \<noteq> []" and "s1 = tgtOf (last (traceOf s al))" and "step s1 a = (ou,s2)"
shows "traceOf s (al ## a) = traceOf s al ## Trans s1 a ou s2"
using assms by (induct al arbitrary: s) (auto split: prod.splits)

lemma sstep_consR[simp]: 
assumes "sstep s al = (oul,s1)" and "step s1 a = (ou,s2)"
shows "sstep s (al ## a) = (oul ## ou, s2)"
using assms by (induct al arbitrary: oul s s1 ou s2) (auto split: prod.splits)

lemma fst_sstep_consR: 
"fst (sstep s (al ## a)) = fst (sstep s al) ## (fst (step (snd (sstep s al)) a))"
by (cases "sstep s al", cases "step (snd (sstep s al)) a") auto

lemma valid_traceOf[simp]: "al \<noteq> [] \<Longrightarrow> valid (traceOf s al)"
proof(induct al arbitrary: s)
  case (Cons a al)
  thus ?case by (cases "al = []") (auto split: prod.splits)
qed auto

lemma validFrom_traceOf[simp]: "validFrom s (traceOf s al)"
by (cases "al = []") auto

inductive reachFrom :: "'state \<Rightarrow> 'state \<Rightarrow> bool"
  for s :: "'state"
where
  Refl[intro]: "reachFrom s s"
| Step: "\<lbrakk>reachFrom s s'; step s' a = (ou, s'')\<rbrakk> \<Longrightarrow> reachFrom s s''"

lemma reachFrom_Step1: "step s a = (ou, s') \<Longrightarrow> reachFrom s s'"
by (auto intro: reachFrom.Step)

lemma reachFrom_Step_Left: "reachFrom s' s'' \<Longrightarrow> step s a = (ou, s') \<Longrightarrow> reachFrom s s''"
by (induction s'' rule: reachFrom.induct) (auto intro: reachFrom.Step)

lemma reachFrom_trans: "reachFrom s0 s1 \<Longrightarrow> reachFrom s1 s2 \<Longrightarrow> reachFrom s0 s2"
by (induction s1 arbitrary: s2 rule: reachFrom.induct) (auto intro: reachFrom_Step_Left)

lemma reachFrom_reach: "reachFrom s s' \<Longrightarrow> reach s \<Longrightarrow> reach s'"
by (induction rule: reachFrom.induct) (auto intro: reach_PairI)

lemma validFrom_traceOf2:
  assumes "validFrom s tr"
  shows "tr = traceOf s (map actOf tr)"
using assms apply (induction tr arbitrary: s) apply (auto split: prod.splits)
by (metis validFrom_def validTrans 
  valid_ConsE actOf.simps fst_conv list.distinct(1) list.sel(1) 
outOf.simps snd_conv srcOf.simps tgtOf.simps validTrans.cases)+

lemma valid_validTrans_set:
assumes "valid tr" and "trn \<in>\<in> tr"
shows "validTrans trn"
using assms by (induct tr arbitrary: trn) auto

lemma validFrom_validTrans_set:
assumes "validFrom s tr" and "trn \<in>\<in> tr"
shows "validTrans trn"
by (metis assms validFrom_def empty_iff list.set valid_validTrans_set)

lemma set_traceOf_validTrans:
assumes "trn \<in>\<in> traceOf s al"  shows "validTrans trn"
by (metis assms validFrom_traceOf validFrom_validTrans_set)

lemma valid_validTrans_nth:
assumes v: "valid tr" and i: "i < length tr"
shows "validTrans (tr!i)"
using valid_validTrans_set[OF v] i by auto

lemma valid_validTrans_nth_srcOf_tgtOf:
assumes v: "valid tr" and i: "Suc i < length tr"
shows "srcOf (tr!(Suc i)) = tgtOf (tr!i)"
by (metis Cons_nth_drop_Suc IO_Automaton.valid_append Suc_lessD append_self_conv2 hd_drop_conv_nth i id_take_nth_drop list.distinct(1) v valid_ConsE)

lemma validFrom_reach: "validFrom s tr \<Longrightarrow> reach s \<Longrightarrow> tr \<noteq> [] \<Longrightarrow> reach (tgtOf (last tr))"
by (intro valid_reach_src_tgt) (auto simp add: validFrom_def)

lemma traceOf_append_sstep: "traceOf s (al @ al1) = traceOf s al @ traceOf (snd (sstep s al)) al1"
by (induction al arbitrary: s al1) (auto split: prod.splits)

lemma snd_sstep_append: "snd (sstep s (al @ al1)) = snd (sstep (snd (sstep s al)) al1)" 
by (cases "sstep s al", cases "sstep (snd (sstep s al)) al1") (auto simp add: sstep_append)

lemma snd_sstep_step_constant: 
assumes "\<forall> a. a \<in>\<in> al \<longrightarrow> snd (step s a) = s"
shows "snd (sstep s al) = s"
using assms by (induction al) (auto split: prod.splits)

lemma traceOf_snoc: 
"traceOf s (al ## a) = 
  traceOf s al ## 
  Trans (snd (sstep s al)) 
        a 
        (fst (step (snd (sstep s al)) a)) 
        (snd (step (snd (sstep s al)) a))"
by (metis (no_types, lifting) traceOf_Nil traceOf_append_sstep prod.case_eq_if traceOf.simps)

lemma traceOf_append_unfold: 
"traceOf s (al1 @ al2) = 
 traceOf s al1 @ traceOf (if al1 = [] then s else tgtOf (last (traceOf s al1))) al2"
using traceOf_append by (cases "al1 = []") auto

abbreviation "transOf s a \<equiv> Trans s a (fst (step s a)) (snd (step s a))"

lemma traceOf_Cons: "traceOf s (a # al) = transOf s a # traceOf (snd (step s a)) al"
by (auto split: prod.splits)

(* The notion of two actions commuting: *)
definition "commute s a1 a2
 \<equiv> snd (sstep s [a1, a2]) = snd (sstep s [a2, a1])"

(* The notion of one action absorbing another: *)
definition absorb :: "'state \<Rightarrow> 'act \<Rightarrow> 'act \<Rightarrow> bool" where 
"absorb s a1 a2 \<equiv> snd (sstep s [a1, a2]) = snd (step s a2)"

lemma validFrom_commute:
  assumes "validFrom s0 (tr1 @ transOf s a # transOf (snd (step s a)) a' # tr2)"
      and "commute s a a'"
  shows "validFrom s0 (tr1 @ transOf s a' # transOf (snd (step s a')) a # tr2)"
using assms unfolding commute_def by (auto split: prod.splits simp add: validFrom_append validFrom_Cons)

lemma validFrom_absorb:
  assumes "validFrom s0 (tr1 @ transOf s a # transOf (snd (step s a)) a' # tr2)"
      and "absorb s a a'"
  shows "validFrom s0 (tr1 @ transOf s a' # tr2)"
using assms unfolding absorb_def by (auto split: prod.splits simp add: validFrom_append validFrom_Cons)

lemma validTrans_Trans_srcOf_actOf_tgtOf: 
"validTrans trn \<Longrightarrow> Trans (srcOf trn) (actOf trn) (outOf trn) (tgtOf trn) = trn"
by (cases trn) auto

lemma validTrans_step_srcOf_actOf_tgtOf: 
"validTrans trn \<Longrightarrow> step (srcOf trn) (actOf trn) = (outOf trn, tgtOf trn)"
by (cases trn) auto

lemma sstep_Cons: 
"sstep s (a # al) = (fst (step s a) # fst (sstep (snd (step s a)) al), snd (sstep (snd (step s a)) al))"
by (auto split: prod.splits)
declare sstep.simps(2)[simp del]

lemma length_fst_sstep: "length (fst (sstep s al)) = length al"
by (induction al arbitrary: s) (auto simp: sstep_Cons)

sublocale TS: Transition_System where istate = istate and validTrans = validTrans
and srcOf = srcOf and tgtOf = tgtOf
done

lemma [simp]: "TS.valid tr \<longleftrightarrow> valid tr"
proof
  assume "TS.valid tr"
  then show "valid tr" by (induction rule: TS.valid.induct) auto
next
  assume "valid tr"
  then show "TS.valid tr" by (induction rule: valid.induct) auto
qed

lemma [simp]: "TS.validFrom s tr \<longleftrightarrow> validFrom s tr"
unfolding TS.validFrom_def validFrom_def by auto

lemma [simp]: "TS.reach s \<longleftrightarrow> reach s"
proof
  assume "TS.reach s"
  then show "reach s" proof (induction rule: TS.reach.induct)
    case (Step s trn s') then show ?case by (cases trn) (auto intro: reach_PairI)
  qed (auto intro: reach.Istate)
next
  assume "reach s"
  then show "TS.reach s" proof (induction rule: reach.induct)
    case (Step s a)
      let ?trn = "Trans s a (fst (step s a)) (snd (step s a))"
      from Step show ?case by (intro TS.reach.Step[of s ?trn "snd (step s a)"]) auto
  qed (auto intro: TS.reach.Istate)
qed

end (* locale IO_Automaton *)






end
